/*
 * pwmh.c
 *
 *  Created on: 20 mai 2021
 *      Author: Florian
 */

/* Local includes */
#include "pwmh.h"

/* Local constants and macros */
#define mPWMH1_SetDYC(dyc)  __HAL_TIM_SET_COMPARE(&htim1, TIM_CHANNEL_1, (uint32_t)dyc)

/* Local typedefs and enums */
typedef enum
{
	cPWMH_STM_RESET, 			// Reset state
	cPWMH_STM_OFF,   			// OFF state - set point is 0
	cPWMH_STM_KICKSTART_RAMP,	// PWM is transitioned from 0 to kickstart threshold
	cPWMH_STM_KICKSTART,		// PWM is maintained at kickstart threshold during kickstart delay
	cPWMH_STM_RAMP, 			// PWM is transitioned to set point
	cPWMH_STM_SET, 				// PWM is fixed, set point is reached
} tPWMH_stm;

/* Local variables */

tPWMH_stm pwm1_stm_T;
uint8_t pwm1_kickstart_threshold_u8;
uint8_t pwm1_dyc_u8, pwm1_dyc_prev_u8;
uint16_t pwm1_ramp_cnt_u16, pwm1_ramp_cumul_u16;
uint16_t pwm1_ramp_time_mem_u16;
uint16_t pwm1_dyc_setpoint_u8;
BOOL pwm1_inc_B;
uint8_t  pwm1_inc_u8;
uint16_t pwm1_time_delta_u16;
uint16_t pwm1_kickstart_delay_u16;
uint16_t pwm1_kickstart_ramp_time_u16;
uint8_t delta_u8;
BOOL new_setpoint_B;

/* Local static functions */

static void StmUpdate (void)
{
	/* PWM1 main state machine*/
	switch (pwm1_stm_T)
	{
	case cPWMH_STM_RESET:
		pwm1_stm_T = cPWMH_STM_OFF;
		break;

	case cPWMH_STM_OFF:
		if (pwm1_dyc_setpoint_u8 > 0U) // PWM is activated
		{
			pwm1_ramp_cnt_u16 = 0U; // Reset timer

			if (pwm1_dyc_setpoint_u8 < pwm1_kickstart_threshold_u8)
			{ // Kickstart is required - Start ramp at kickstart threshold
				pwm1_stm_T = cPWMH_STM_KICKSTART_RAMP;
				pwm1_inc_B = FALSE; // Set point is lower than the kickstart threshold,
									//   thus the ramp after the kickstart will be downhill
				delta_u8 = pwm1_kickstart_threshold_u8 - pwm1_dyc_setpoint_u8;
				if (pwm1_kickstart_ramp_time_u16 > pwm1_kickstart_threshold_u8)
				{
					pwm1_ramp_cumul_u16 = pwm1_ramp_time_mem_u16 >> 1U;
				}
				else
				{
					pwm1_ramp_cumul_u16 = delta_u8 >> 1U;
				}
			}
			else
			{ // No kickstart required - Start ramp at zero
				pwm1_stm_T = cPWMH_STM_RAMP;
				pwm1_dyc_prev_u8 = 0U;
				pwm1_inc_B = TRUE; // Start is 0, thus ramp is uphill
				delta_u8 = pwm1_dyc_setpoint_u8;

				if (pwm1_ramp_time_mem_u16 > delta_u8)
				{
					pwm1_ramp_cumul_u16 = pwm1_ramp_time_mem_u16 >> 1U;
				}
				else
				{
					pwm1_ramp_cumul_u16 = delta_u8 >> 1U;
				}
			}
		}
		break;

	case cPWMH_STM_KICKSTART_RAMP:
		if (pwm1_dyc_u8 >= pwm1_kickstart_threshold_u8)
		{
			pwm1_ramp_cnt_u16 = 0U; // Reset timer
			pwm1_dyc_u8 = pwm1_kickstart_threshold_u8;
			pwm1_stm_T = cPWMH_STM_KICKSTART;
		}
		else
		{ // Continue Bresenham algo
			if (pwm1_kickstart_ramp_time_u16 > pwm1_kickstart_threshold_u8)
			{ // More delay than DYC delta : wait before incrementing
				//pwm1_ramp_cumul_u16 = pwm1_kickstart_delay_u16 >> 1U;

				if (pwm1_ramp_cnt_u16 < pwm1_kickstart_ramp_time_u16)
				{
					pwm1_ramp_cnt_u16++;
					pwm1_ramp_cumul_u16 += pwm1_kickstart_threshold_u8;

					if (pwm1_ramp_cumul_u16 > pwm1_kickstart_ramp_time_u16)
					{ // Time to increment the PWM
						pwm1_ramp_cumul_u16 -= pwm1_kickstart_ramp_time_u16;
						pwm1_dyc_u8++;
					}
				}
			}
			else
			{ // More DYC delta than delay : increment at each pass
				//pwm1_ramp_cumul_u16 = pwm1_kickstart_threshold_u8 >> 1U;

				while (pwm1_ramp_cnt_u16 < pwm1_kickstart_threshold_u8)
				{
					pwm1_ramp_cnt_u16++;
					pwm1_ramp_cumul_u16 += pwm1_kickstart_ramp_time_u16;

					if (pwm1_ramp_cumul_u16 > pwm1_kickstart_threshold_u8)
					{ // Step is applied
						pwm1_ramp_cumul_u16 -= pwm1_kickstart_threshold_u8;
						pwm1_dyc_u8 += pwm1_ramp_cnt_u16;

						if (pwm1_dyc_u8 > pwm1_kickstart_threshold_u8)
						{
							pwm1_dyc_u8 = pwm1_kickstart_threshold_u8;
						}
						break;
					}
				}
			}
		}
		break;

	case cPWMH_STM_KICKSTART:
		pwm1_dyc_u8 = pwm1_kickstart_threshold_u8;
		pwm1_ramp_cnt_u16++;

		// Check eventual set point update

		if (pwm1_dyc_setpoint_u8 < pwm1_kickstart_threshold_u8)
		{
			pwm1_inc_B = FALSE; // Set point is lower than the kickstart threshold,
								//   thus the ramp after the kickstart will be downhill
			delta_u8 = pwm1_kickstart_threshold_u8 - pwm1_dyc_setpoint_u8;
			pwm1_dyc_prev_u8 = pwm1_kickstart_threshold_u8;
		}
		else
		{ // No kickstart required - Start ramp at zero
			pwm1_dyc_prev_u8 = pwm1_kickstart_threshold_u8;
			pwm1_inc_B = TRUE; // Start is 0, thus ramp is uphill
			delta_u8 = pwm1_dyc_setpoint_u8 - pwm1_kickstart_threshold_u8;
		}

		if (pwm1_ramp_cnt_u16 >= pwm1_kickstart_delay_u16)
		{
			pwm1_ramp_cnt_u16 = 0U; // Reset timer
			pwm1_dyc_prev_u8 = pwm1_dyc_u8; // Latch start value
			if (pwm1_ramp_time_mem_u16 > delta_u8)
			{
				pwm1_ramp_cumul_u16 = pwm1_ramp_time_mem_u16 >> 1U;
			}
			else
			{
				pwm1_ramp_cumul_u16 = delta_u8 >> 1U;
			}
			pwm1_stm_T = cPWMH_STM_RAMP;
		}
		break;

	case cPWMH_STM_RAMP:
		if (pwm1_dyc_u8 == pwm1_dyc_setpoint_u8)
		{
			pwm1_ramp_cnt_u16 = 0U; // Reset timer
			pwm1_stm_T = cPWMH_STM_SET;
		}
		else
		{ // Continue Bresenham algo
			if (new_setpoint_B == TRUE)
			{ // If incoming new values
				new_setpoint_B = FALSE; // Reset flag
				pwm1_ramp_cnt_u16 = 0U; // Reset timer
				pwm1_dyc_prev_u8 = pwm1_dyc_u8; // Latch start value
				if (pwm1_ramp_time_mem_u16 > delta_u8)
				{
					pwm1_ramp_cumul_u16 = pwm1_ramp_time_mem_u16 >> 1U;
				}
				else
				{
					pwm1_ramp_cumul_u16 = delta_u8 >> 1U;
				}
			}

			if (pwm1_ramp_time_mem_u16 > delta_u8)
			{ // More delay than DYC delta : wait before incrementing
				//pwm1_ramp_cumul_u16 = pwm1_ramp_time_mem_u16 >> 1U;

				if (pwm1_ramp_cnt_u16 < pwm1_ramp_time_mem_u16)
				{
					pwm1_ramp_cnt_u16++;
					pwm1_ramp_cumul_u16 += delta_u8;

					if (pwm1_ramp_cumul_u16 > pwm1_ramp_time_mem_u16)
					{ // Time to increment the PWM
						pwm1_ramp_cumul_u16 -= pwm1_ramp_time_mem_u16;

						if (pwm1_inc_B == TRUE)
						{
							pwm1_dyc_u8++;
						}
						else
						{
							pwm1_dyc_u8--;
						}
					}
				}
			}
			else
			{ // More DYC delta than delay : increment at each pass
				//pwm1_ramp_cumul_u16 = delta_u8 >> 1U;

				while (pwm1_ramp_cnt_u16 < delta_u8)
				{
					pwm1_ramp_cnt_u16++;
					pwm1_ramp_cumul_u16 += pwm1_ramp_time_mem_u16;

					if (pwm1_ramp_cumul_u16 > delta_u8)
					{ // Step is applied
						pwm1_ramp_cumul_u16 -= delta_u8;

						if (pwm1_inc_B == TRUE)
						{
							pwm1_dyc_u8 += pwm1_ramp_cnt_u16;
							if (pwm1_dyc_u8 > pwm1_dyc_setpoint_u8)
							{
								pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
							}
						}
						else
						{
							pwm1_dyc_u8 -= pwm1_ramp_cnt_u16;
							if (pwm1_dyc_u8 < pwm1_dyc_setpoint_u8)
							{
								pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
							}
						}
						break;
					}
				}
			}
		}
		break;

	case cPWMH_STM_SET:
		if (pwm1_dyc_setpoint_u8 != pwm1_dyc_u8) // PWM is changed
		{
			pwm1_ramp_cnt_u16 = 0U; // Reset timer
			pwm1_dyc_prev_u8 = pwm1_dyc_u8; // Latch start value

			if (pwm1_ramp_time_mem_u16 > delta_u8)
			{
				pwm1_ramp_cumul_u16 = pwm1_ramp_time_mem_u16 >> 1U;
			}
			else
			{
				pwm1_ramp_cumul_u16 = delta_u8 >> 1U;
			}

			if (pwm1_dyc_u8 < pwm1_dyc_setpoint_u8)
			{
				pwm1_inc_B = TRUE;
				delta_u8 = pwm1_dyc_setpoint_u8 - pwm1_dyc_prev_u8;
			}
			else
			{
				pwm1_inc_B = FALSE;
				delta_u8 = pwm1_dyc_prev_u8 - pwm1_dyc_setpoint_u8;
			}

			pwm1_stm_T = cPWMH_STM_RAMP;
		}
		break;

	default:
		break;
	}
}

/* Shared functions */

void PWMH_Init (void)
{
	MX_TIM1_Init();

	/* Init PWM1 */
	pwm1_stm_T = cPWMH_STM_RESET;
	pwm1_dyc_setpoint_u8 = 0U;
	pwm1_dyc_u8 = 0U;
	pwm1_dyc_prev_u8 = 0U;
	pwm1_ramp_cnt_u16 = 0U;
	pwm1_ramp_cumul_u16 = 0U;
	pwm1_ramp_time_mem_u16 = 0U;
	pwm1_inc_B = TRUE;
	pwm1_inc_u8 = 0U;
	pwm1_time_delta_u16 = 0U;
	pwm1_kickstart_threshold_u8 = cPWMH_KICKSTART_THRESHOLD_DEFAULT_U8;
	pwm1_kickstart_delay_u16 = cPWMH_KICKSTART_DELAY_DEFAULT_U16;
	pwm1_kickstart_ramp_time_u16 = cPWMH_KICKSTART_RAMP_TIME_DEFAULT_U16;
	delta_u8 = 0U;
	new_setpoint_B = FALSE;
}

void PWMH_Start (void)
{
	mPWMH1_SetDYC(pwm1_dyc_u8);
	HAL_TIM_PWM_Start_IT(&htim1, TIM_CHANNEL_1);
}

void PWMH_Stop (void)
{

}

void PWMH_PeriodicFunction (void)
{


	StmUpdate();

	/*if (pwm1_dyc_u8 != pwm1_dyc_setpoint_u8) // Setpoint not reached - Ramp ongoing
	{
		if (pwm1_ramp_cnt_u16 > pwm1_time_delta_u16) // Is it time to increment / decrement ?
		{
			pwm1_ramp_cnt_u16 = 0U;
			if (pwm1_inc_B == TRUE) // Increment / rising ramp
			{
				if (pwm1_dyc_u8 <= ((uint8_t)255 - pwm1_inc_u8)) // Cap at 255
				{
					pwm1_dyc_u8 += pwm1_inc_u8;
					if (pwm1_dyc_u8 > pwm1_dyc_setpoint_u8) // If setpoint passed over -> cap at setpoint
					{
						pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
					}
				}
				else // If 255 reached while searching for setpoint -> setpoint is passed -> cap at setpoint
				{
					pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
				}
			}
			else // Decrement / falling ramp
			{
				if (pwm1_dyc_u8 >= pwm1_inc_u8) // Cap at 0
				{
					pwm1_dyc_u8 -= pwm1_inc_u8;
					if (pwm1_dyc_u8 < pwm1_dyc_setpoint_u8) // If setpoint passed over -> cap at setpoint
					{
						pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
					}
				}
				else // If 0 reached while searching for setpoint -> setpoint is passed -> cap at setpoint
				{
					pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
				}
			}
		}
	}*/

	mPWMH1_SetDYC(pwm1_dyc_u8); // Apply new duty cycle value
}

BOOL PWMH_PWM1SetRamp (uint8_t dyc_u8, uint16_t ramp_time_u16)
{
	BOOL ret_B = FALSE;

	new_setpoint_B = TRUE;

	if (dyc_u8 != pwm1_dyc_u8)
	{
		pwm1_dyc_setpoint_u8 = dyc_u8;

		if (ramp_time_u16 == 0U) // Ramp time 0 : apply DYC instantly
		{
			pwm1_ramp_time_mem_u16 = ramp_time_u16;
			pwm1_dyc_u8 = pwm1_dyc_setpoint_u8;
			ret_B = TRUE;
			return ret_B;
		}

		pwm1_ramp_time_mem_u16 = ramp_time_u16;

		if (dyc_u8 > pwm1_dyc_u8)
		{
			pwm1_inc_B = TRUE;
			delta_u8 = pwm1_dyc_setpoint_u8 - pwm1_dyc_u8;
		}
		else
		{
			pwm1_inc_B = FALSE;
			delta_u8 = pwm1_dyc_u8 - pwm1_dyc_setpoint_u8;
		}
	}


	return ret_B;
}
